from ..schema import FeatureSchema, Schema, SchemaDerivation, VersionedId
from ..transformations import Transformation
from .base_job import BaseJob


class ExtractSchemaJob(BaseJob):
    def _response_method(self, job_id):
        response = self._stub.GetExtractSchemaJobOutput(job_id)
        derivation = None
        if response.derived_from.HasField("transformation"):
            derivation = SchemaDerivation(
                [
                    VersionedId(f.id, f.major_version)
                    for f in response.derived_from.feature_set_ids
                ],
                Transformation.from_proto(response.derived_from.transformation),
            )
        return Schema(
            ExtractSchemaJob._features_schema_from_proto(response.schema),
            True,
            derivation,
        )

    @staticmethod
    def _features_schema_from_proto(schema):
        return [
            FeatureSchema(
                feature_schema.name,
                feature_schema.data_type,
                nested_features_schema=ExtractSchemaJob._features_schema_from_proto(
                    feature_schema.nested
                ),
            )
            for feature_schema in schema
        ]



from ..base import NumericField, StringField


class FeatureProfile(object):
    feature_type = StringField("features.profile.featureType")


class Feature(object):
    name = StringField("features.fullyQualifiedFeatureNames")
    description = StringField("features.description")
    importance = NumericField("features.importance")
    data_type = StringField("features.dataType")
    profile = FeatureProfile



import re
from typing import List

from ai.h2o.featurestore.api.v1 import CoreService_pb2 as pb
from ai.h2o.featurestore.api.v1 import FeatureSetProtoApi_pb2 as FeatureSetApi

from .transformations import Transformation


class FeatureSchemaSpecialData:
    def __init__(
        self, spi=False, pci=False, rpi=False, demographic=False, sensitive=False
    ):
        self.spi = spi
        self.pci = pci
        self.rpi = rpi
        self.demographic = demographic
        self.sensitive = sensitive

    def __repr__(self):
        return f"spi={self.spi}, pci={self.pci}, rpi={self.rpi}, demographic={self.demographic}, sensitive={self.sensitive}"


class FeatureSchema:
    def __init__(
        self,
        name,
        _data_type,
        special_data=None,
        nested_features_schema=None,
        _feature_type="AUTOMATIC_DISCOVERY",
        description="",
        classifiers: set = None,
    ):
        self._data_type = _data_type
        self._name = name
        self._special_data = special_data or FeatureSchemaSpecialData()
        self._nested_features_schema = nested_features_schema or []
        self._schema = Schema(self._nested_features_schema, False)
        self._validated_feature_type = FeatureSchema.get_and_validate_feature_type(
            _feature_type
        )
        self._description = description
        self._classifiers = classifiers or set()

    @property
    def schema(self):
        return self._schema

    @property
    def special_data(self):
        return self._special_data

    @property
    def name(self):
        return self._name

    @property
    def data_type(self):
        if self._schema.size() > 0:
            return self._data_type.format(self._schema.to_string())
        else:
            return self._data_type

    @data_type.setter
    def data_type(self, value):
        self._data_type = value

    @property
    def feature_type(self):
        return FeatureSetApi.FeatureType.Name(self._validated_feature_type)

    @feature_type.setter
    def feature_type(self, value):
        self._validated_feature_type = FeatureSchema.get_and_validate_feature_type(
            value
        )

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def classifiers(self):
        return self._classifiers

    @classifiers.setter
    def classifiers(self, value: set):
        self._classifiers = value

    @property
    def schema(self):
        return self._schema

    def to_string(self):
        return self.data_type

    @staticmethod
    def create_from(feature):
        nested = [
            FeatureSchema.create_from(nested_feature)
            for _, nested_feature in feature.nested_features.items()
        ]
        return FeatureSchema(
            feature.name,
            feature.data_type,
            FeatureSchemaSpecialData(
                feature.special_data.spi,
                feature.special_data.pci,
                feature.special_data.rpi,
                feature.special_data.demographic,
                feature.special_data.sensitive,
            ),
            nested,
            feature.profile.feature_type,
            feature.description,
            feature.classifiers,
        )

    @staticmethod
    def get_and_validate_feature_type(feature_type):
        valid_values = FeatureSetApi.FeatureType.keys()
        if feature_type.upper() in valid_values:
            FeatureSetApi.FeatureType.Value(feature_type.upper())
        else:
            raise Exception(
                "Invalid feature type. Supported values are: "
                + ", ".join(map(str, valid_values))
            )

    def __repr__(self):
        return self.to_string()


class VersionedId:
    def __init__(self, id: str, major_version: int):
        self.id = id
        self.major_version = major_version


class SchemaDerivation:
    def __init__(
        self,
        feature_set_ids: List[VersionedId],
        transformation: Transformation,
    ):
        self.feature_set_ids = feature_set_ids
        self.transformation = transformation

    def _to_proto(self):
        return pb.DerivedInformation(
            feature_set_ids=[
                pb.VersionedId(id=f.id, major_version=f.major_version)
                for f in self.feature_set_ids
            ],
            transformation=self.transformation._to_proto(),
        )


class Schema:
    def __init__(
        self, feature_schemas, is_root_schema, derivation: SchemaDerivation = None
    ):
        self._feature_schemas = feature_schemas
        self._is_root_schema = is_root_schema
        if is_root_schema:
            self._field_separator = " "
        else:
            self._field_separator = " : "
        self.derivation = derivation

    def __getitem__(self, feature_name):
        filtered = list(filter(lambda f: f.name == feature_name, self._feature_schemas))
        if len(filtered):
            return filtered[0]
        else:
            raise Exception(
                f"Feature name '{feature_name}' does not exist in the schema."
            )

    @staticmethod
    def create_from(obj):
        from .entities.feature_set import FeatureSet

        if isinstance(obj, FeatureSet):
            features = [
                FeatureSchema.create_from(feature)
                for name, feature in obj.features.items()
            ]
            if obj.is_derived():
                derived_from = obj._feature_set.derived_from
                derivation = SchemaDerivation(
                    [
                        VersionedId(f.id, f.major_version)
                        for f in derived_from.feature_set_ids
                    ],
                    Transformation.from_proto(derived_from.transformation),
                )
                return Schema(features, True, derivation)
            else:
                return Schema(features, True)
        elif isinstance(obj, str):
            return Schema(Schema._parse_features_schema(obj, True)[::-1], True)
        else:
            raise Exception(
                "Schema can be created either from string or existing feature set."
            )

    @staticmethod
    def create_derived_from(
        schema_string: str, derive_from, transformation: Transformation
    ):
        return Schema(
            Schema._parse_features_schema(schema_string, True)[::-1],
            True,
            SchemaDerivation(
                [VersionedId(f.id, f.major_version) for f in derive_from],
                transformation,
            ),
        )

    def select(self, feature_names):
        self._feature_schemas = list(
            filter(lambda schema: schema.name in feature_names, self._feature_schemas)
        )
        return self

    def exclude(self, feature_names):
        self._feature_schemas = list(
            filter(
                lambda schema: schema.name not in feature_names, self._feature_schemas
            )
        )
        return self

    def append(self, feature_schema, after=None):
        if not after:
            self._feature_schemas.append(feature_schema)
        else:
            index = self._feature_schemas.index(after) + 1
            self._feature_schemas.insert(index, feature_schema)
        return self

    def prepend(self, feature_schema, before=None):
        if not before:
            self._feature_schemas.insert(0, feature_schema)
        else:
            index = self._feature_schemas.index(before)
            self._feature_schemas.insert(index, feature_schema)
        return self

    def size(self):
        return len(self._feature_schemas)

    def is_derived(self):
        return self.derivation is not None

    def to_string(self):
        return ", ".join(
            [
                f"{feature_schema.name}{self._field_separator}{feature_schema.data_type}"
                for feature_schema in self._feature_schemas
            ]
        )

    def _to_proto_schema(self):
        return [
            pb.FeatureSchema(
                name=feature_schema.name,
                data_type=feature_schema.data_type,
                nested=feature_schema.schema._to_proto_schema(),
                special_data=pb.FeatureSchema.SpecialData(
                    spi=feature_schema.special_data.spi,
                    pci=feature_schema.special_data.pci,
                    rpi=feature_schema.special_data.rpi,
                    demographic=feature_schema.special_data.demographic,
                    sensitive=feature_schema.special_data.sensitive,
                ),
                feature_type=feature_schema._validated_feature_type,
                description=feature_schema.description,
                classifiers=list(feature_schema.classifiers),
            )
            for feature_schema in self._feature_schemas
        ]

    def __repr__(self):
        return self.to_string()

    @staticmethod
    def _parse_features_schema(schema_string, is_root_level):
        if schema_string:
            feature_schema, remaining_schema_string = Schema._extract_pair(
                schema_string, is_root_level
            )
            return Schema._parse_features_schema(
                remaining_schema_string, is_root_level
            ) + [feature_schema]
        else:
            return []

    @staticmethod
    def _get_index_of_matching_ending_angular_bracket(string):
        stack = []
        for idx, ch in enumerate(string):
            if ch == "<":
                stack.append(idx)
            elif ch == ">":
                if len(stack) == 1:
                    return idx
                else:
                    stack.pop()

    @staticmethod
    def _extract_pair(schema, is_root_level):
        if is_root_level:
            splits = schema.split()
            remaining_with_data_type = " ".join(splits[1:])
        else:
            splits = re.split(r"\s*:\s*", schema)
            remaining_with_data_type = " : ".join(splits[1:])
        col_name = splits[0]
        placeholder = None
        if Schema._is_complex(remaining_with_data_type):
            idx = (
                Schema._get_index_of_matching_ending_angular_bracket(
                    remaining_with_data_type
                )
                + 1
            )
            nested_data_type, remaining_schema = (
                remaining_with_data_type[: idx + 1],
                remaining_with_data_type[idx + 1 :],
            )
            if remaining_with_data_type.upper().startswith("STRUCT"):
                nested_features_schema, placeholder = Schema._extract_struct(
                    nested_data_type
                )
            else:
                nested_features_schema, placeholder = Schema._extract_array(
                    nested_data_type
                )
            if remaining_schema:
                data_type = placeholder or nested_data_type
                rest = " ".join(
                    [
                        split.strip()
                        for split in re.split(r",<?![^<]*>>", remaining_schema)
                    ]
                )
                nested_schema = nested_features_schema
            else:
                data_type = placeholder or nested_data_type
                rest = ""
                nested_schema = nested_features_schema
        else:
            data_type, rest, nested_schema = Schema._extract_primitive(
                remaining_with_data_type
            )
        return (
            FeatureSchema(col_name, data_type, nested_features_schema=nested_schema),
            rest,
        )

    @staticmethod
    def _is_complex(data_type):
        first_non_array_type = Schema._extract_first_non_array_data_type(data_type)
        return first_non_array_type.upper().startswith("STRUCT")

    @staticmethod
    def _extract_first_non_array_data_type(data_type):
        if data_type.upper().startswith("ARRAY"):
            nested_data_type = re.split(
                r"\s+<", data_type[len("ARRAY") + 1 : len(data_type) - 1]
            )[-1].strip()
            return Schema._extract_first_non_array_data_type(nested_data_type)
        else:
            return data_type

    @staticmethod
    def _extract_struct(data_type):
        nested_data_type = re.split(
            r"\s+<", data_type[len("STRUCT") : len(data_type) - 1].strip("<>")
        )[-1].strip()
        return (
            Schema._parse_features_schema(nested_data_type, is_root_level=False)[::-1],
            "STRUCT<{0}>",
        )

    @staticmethod
    def _extract_array(data_type):
        nested_data_type = re.split(
            r"\s+<", data_type[len("ARRAY") + 1 : len(data_type) - 1]
        )[-1].strip()
        if nested_data_type.upper().startswith("ARRAY"):
            return Schema._extract_array(nested_data_type)
        elif nested_data_type.upper().startswith("STRUCT"):
            inner, placeholder = Schema._extract_struct(nested_data_type)
            return inner, f"ARRAY<{placeholder}>"
        else:
            return [], ""

    @staticmethod
    def _extract_primitive(schema):
        splits = [split.strip() for split in schema.split(",")]
        if len(splits) > 1:
            return splits[0], ", ".join(splits[1:]), []
        else:
            return splits[0], "", []



import warnings

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
import ai.h2o.featurestore.api.v1.FeatureSetSearch_pb2 as pb_search
from ai.h2o.featurestore.api.v1.FeatureSetSearch_pb2 import (
    BooleanFilter,
    NumericalFilter,
    TextualFilter,
)

from .commons.spark_utils import SparkUtils


class DeltaTableFilter:
    def __init__(self, column, operator, value):
        self.column = column
        self.operator = operator
        self.value = value
        self._filter = self.__build()

    def __build(self):
        if isinstance(self.value, str):
            return pb_search.Filter(
                text=TextualFilter(
                    field=self.column, operator=self.operator, value=[self.value]
                )
            )
        elif isinstance(self.value, (int, float)):
            return pb_search.Filter(
                numeric=NumericalFilter(
                    field=self.column, operator=self.operator, value=self.value
                )
            )
        elif isinstance(self.value, bool):
            return pb_search.Filter(
                boolean=BooleanFilter(
                    field=self.column, operator=self.operator, value=self.value
                )
            )


class SparkDataFrame:
    def __init__(self, dataframe):
        self.dataframe = dataframe
        self._write_path = None

    def _write_to_cache(self, stub):
        from pyspark.sql import SparkSession  # Local import

        spark = SparkSession.builder.getOrCreate()

        session_id = spark.conf.get("ai.h2o.featurestore.sessionId", "")
        request = pb.IngestWriteCredentialRequest()
        request.session_id = session_id
        write_info = stub.GetIngestWriteCredentials(request)

        SparkUtils.configure_user_spark(spark)
        spark.conf.set("ai.h2o.featurestore.sessionId", write_info.session_id)
        for k, v in write_info.options.items():
            spark.conf.set(k, v)
            if k.startswith("spark.hadoop."):
                spark.sparkContext._jsc.hadoopConfiguration().set(
                    k.replace("spark.hadoop.", ""), v
                )
        spark.conf.set("spark.sql.session.timeZone", "UTC")
        self.dataframe.write.parquet(write_info.path)
        self._write_path = write_info.path

    def _get_cache_location(self):
        if not self._write_path:
            raise Exception(
                "Write the spark frame into the cache before calling get method"
            )
        parquet = pb.TempParquetFileSpec()
        parquet.path = self._write_path
        raw_data_location = pb.RawDataLocation()
        raw_data_location.temp_parquet.CopyFrom(parquet)
        return raw_data_location


class CSVFile:
    def __init__(self, path, delimiter=","):
        self.path = path
        self.delimiter = delimiter


class JSONFile:
    def __init__(self, path, multiline=False):
        self.path = path
        self.multiline = multiline


class ParquetFile:
    def __init__(self, path):
        self.path = path


class Proxy:
    def __init__(self, host="", port=0, user="", password=""):
        if port and not host:
            raise ValueError(f"Proxy port specified but host is missing!")
        if not port and host:
            raise ValueError(f"Proxy host specified but port is missing!")
        if port and host:
            if user and not password:
                raise ValueError(f"Proxy user specified but password is missing!")
            if not user and password:
                raise ValueError(f"Proxy password specified but user is missing!")
        self.host = host
        self.port = port
        self.user = user
        self.password = password


class SnowflakeTable:
    def __init__(
        self,
        url,
        warehouse,
        database,
        schema,
        table="",
        query="",
        insecure=False,
        proxy=None,
        role="",
        account="",
    ):
        if not (query or table):
            raise ValueError(f"table or query is required!")
        if query and table:
            raise ValueError(f"Only one of table or query is supported!")

        self.table = table
        self.database = database
        self.url = url
        self.query = query
        self.warehouse = warehouse
        self.schema = schema
        self.insecure = insecure
        self.proxy = proxy
        self.role = role
        self.account = account


class JdbcTable:
    def __init__(
        self,
        connection_url,
        table="",
        query="",
        partition_options=None,
    ):
        if not (table or query):
            raise ValueError("Table or query is required!")
        if table and query:
            raise ValueError("Only one of table or query is supported!")
        self.table = table
        self.query = query
        self.connection_url = connection_url
        self.partition_options = partition_options


class PartitionOptions:
    def __init__(
        self,
        num_partitions=None,
        partition_column=None,
        lower_bound=None,
        upper_bound=None,
        fetch_size=1000,
    ):
        self.num_partitions = num_partitions
        self.partition_column = partition_column
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.fetch_size = fetch_size


class SnowflakeCursor:
    def __init__(
        self,
        url,
        warehouse,
        database,
        schema,
        cursor,
        insecure=False,
        proxy=None,
        role="",
        account="",
    ):
        self.cursor = cursor
        self.database = database
        self.url = url
        self.warehouse = warehouse
        self.schema = schema
        self.insecure = insecure
        self.proxy = proxy
        self.role = role
        self.account = account

    def get_latest_query(self):
        query = "SELECT QUERY_TEXT::VARCHAR FROM TABLE(INFORMATION_SCHEMA.QUERY_HISTORY_BY_SESSION(RESULT_LIMIT => 10)) WHERE QUERY_ID=%s"
        self.cursor.execute(
            query, (self.cursor.sfqid,)
        )  # get the last executed query using the query id
        try:
            latest_query = self.cursor.fetchone()[0]
            if not latest_query.lower().startswith("select "):
                raise ValueError(
                    "Only select queries are supported for registering featuring sets"
                )
        except IndexError:
            raise ValueError("No query seems to have been executed in this session")
        return latest_query


class DeltaTable:
    def __init__(self, path, version=-1, timestamp=None, filter=None):
        self.path = path
        self.version = version
        self.timestamp = timestamp
        self.filter = filter
        if version and timestamp:
            raise ValueError("Only one of version or timestamp is supported")


class CSVFolder:
    def __init__(self, root_folder, delimiter=",", filter_pattern=""):
        self.root_folder = root_folder
        self.filter_pattern = filter_pattern
        self.delimiter = delimiter


class ParquetFolder:
    def __init__(self, root_folder, filter_pattern=""):
        self.root_folder = root_folder
        self.filter_pattern = filter_pattern


class JSONFolder:
    def __init__(self, root_folder, multiline=False, filter_pattern=""):
        self.root_folder = root_folder
        self.multiline = multiline
        self.filter_pattern = filter_pattern


def get_raw_data_location(source):
    raw_data_location = pb.RawDataLocation()
    if isinstance(source, CSVFile):
        csv = pb.CSVFileSpec()
        csv.path = source.path
        csv.delimiter = source.delimiter
        raw_data_location.csv.CopyFrom(csv)
    elif isinstance(source, JSONFile):
        json = pb.JSONFileSpec()
        json.path = source.path
        json.multiline = source.multiline
        raw_data_location.json.CopyFrom(json)
    elif isinstance(source, ParquetFile):
        parquet = pb.ParquetFileSpec()
        parquet.path = source.path
        raw_data_location.parquet.CopyFrom(parquet)
    elif isinstance(source, SnowflakeTable):
        snowflake = pb.SnowflakeTableSpec()
        snowflake.table = source.table
        snowflake.database = source.database
        snowflake.url = source.url
        snowflake.warehouse = source.warehouse
        snowflake.schema = source.schema
        snowflake.query = source.query
        snowflake.insecure = source.insecure
        snowflake.role = source.role
        snowflake.account = source.account
        if source.proxy:
            snowflake.proxy.host = source.proxy.host
            snowflake.proxy.port = source.proxy.port
            snowflake.proxy.user = source.proxy.user
            snowflake.proxy.password = source.proxy.password
        raw_data_location.snowflake.CopyFrom(snowflake)
    elif isinstance(source, SnowflakeCursor):
        snowflake = pb.SnowflakeTableSpec()
        snowflake.table = ""
        snowflake.database = source.database
        snowflake.url = source.url
        snowflake.warehouse = source.warehouse
        snowflake.schema = source.schema
        snowflake.query = source.get_latest_query()
        snowflake.insecure = source.insecure
        snowflake.role = source.role
        if source.proxy:
            snowflake.proxy.host = source.proxy.host
            snowflake.proxy.port = source.proxy.port
            snowflake.proxy.user = source.proxy.user
            snowflake.proxy.password = source.proxy.password
        raw_data_location.snowflake.CopyFrom(snowflake)
    elif isinstance(source, JdbcTable):
        jdbc = pb.JDBCTableSpec()
        jdbc.table = source.table
        jdbc.connection_url = source.connection_url
        jdbc.query = source.query
        if source.partition_options is not None:
            jdbc.num_partitions = source.partition_options.num_partitions
            jdbc.partition_column = source.partition_options.partition_column
            jdbc.lower_bound = source.partition_options.lower_bound
            jdbc.upper_bound = source.partition_options.upper_bound
            jdbc.fetch_size = source.partition_options.fetch_size
        raw_data_location.jdbc.CopyFrom(jdbc)
    elif isinstance(source, DeltaTable):
        delta_table = pb.DeltaTableSpec()
        delta_table.path = source.path
        delta_table.version = source.version
        if source.timestamp:
            delta_table.timestamp = source.timestamp
        if source.filter:
            delta_table.filter.CopyFrom(source.filter._filter)
        raw_data_location.delta_table.CopyFrom(delta_table)
    elif isinstance(source, CSVFolder):
        csv_folder = pb.CSVFolderSpec()
        csv_folder.root_folder = source.root_folder
        csv_folder.filter_pattern = source.filter_pattern
        csv_folder.delimiter = source.delimiter
        raw_data_location.csv_folder.CopyFrom(csv_folder)
    elif isinstance(source, ParquetFolder):
        parquet_folder = pb.ParquetFolderSpec()
        parquet_folder.root_folder = source.root_folder
        parquet_folder.filter_pattern = source.filter_pattern
        raw_data_location.parquet_folder.CopyFrom(parquet_folder)
    elif isinstance(source, JSONFolder):
        json_folder = pb.JSONFolderSpec()
        json_folder.root_folder = source.root_folder
        json_folder.filter_pattern = source.filter_pattern
        json_folder.multiline = source.multiline
        raw_data_location.json_folder.CopyFrom(json_folder)
    else:
        raise Exception("Unsupported external data source.")

    return raw_data_location



from google.protobuf.json_format import Parse

import ai.h2o.featurestore.api.v1.FeatureSetSearch_pb2 as pb


class Field(object):
    def __init__(self, name):
        self.name = name

    def __eq__(self, value):
        return self._build("eq", value)

    def __ne__(self, value):
        return self._build("ne", value)


class StringField(Field):
    def like(self, value):
        return self._build("regex", value)

    def in_(self, *args):
        return self._build("in", args)

    def _build(self, operator, value):
        _filter = pb.Filter()
        text_filter = pb.TextualFilter()
        text_filter.field = self.name
        if isinstance(value, (list, tuple, set)):
            text_filter.value.extend([str(i) for i in value])
        else:
            text_filter.value.append(str(value))
        text_filter.operator = operator
        _filter.text.CopyFrom(text_filter)
        return _filter


class BooleanField(Field):
    def _build(self, operator, value):
        _filter = pb.Filter()
        boolean_filter = pb.BooleanFilter()
        boolean_filter.field = self.name
        boolean_filter.value = bool(value)
        boolean_filter.operator = operator
        _filter.boolean.CopyFrom(boolean_filter)
        return _filter


class NumericField(Field):
    def __init__(self, name):
        super(NumericField, self).__init__(name)
        self._aggregate = None

    def __gt__(self, value):
        return self._build("gt", value)

    def __lt__(self, value):
        return self._build("lt", value)

    def __ge__(self, value):
        return self._build("gte", value)

    def __le__(self, value):
        return self._build("lte", value)

    def sum(self):
        self._aggregate = "sum"
        return self

    def avg(self):
        self._aggregate = "avg"
        return self

    def min(self):
        self._aggregate = "min"
        return self

    def max(self):
        self._aggregate = "max"
        return self

    def _build(self, operator, value):
        _filter = pb.Filter()
        numeric_filter = pb.NumericalFilter()
        numeric_filter.field = self.name
        numeric_filter.value = value
        numeric_filter.operator = operator
        if self._aggregate:
            numeric_filter.aggregate = self._aggregate
        _filter.numeric.CopyFrom(numeric_filter)
        self._aggregate = ""  # reset
        return _filter


class ArrayField(StringField):
    def __init__(self, name):
        super(ArrayField, self).__init__(name)
        self._aggregate = None

    def in_(self, *args):
        return super(ArrayField, self)._build("in", args)

    def length(self):
        self._aggregate = "length"
        return self

    def __gt__(self, value):
        return self._build("gt", value)

    def __lt__(self, value):
        return self._build("lt", value)

    def __ge__(self, value):
        return self._build("gte", value)

    def __le__(self, value):
        return self._build("lte", value)

    def _build(self, operator, value):
        _filter = pb.Filter()
        if isinstance(value, str):
            text_filter = pb.TextualFilter()
            text_filter.field = self.name
            text_filter.value.extend([value])
            text_filter.operator = operator
            _filter.text.CopyFrom(text_filter)
        else:
            numeric_filter = pb.NumericalFilter()
            numeric_filter.field = self.name
            numeric_filter.value = value
            numeric_filter.operator = operator
            if self._aggregate:
                numeric_filter.aggregate = self._aggregate
            _filter.numeric.CopyFrom(numeric_filter)
        self._aggregate = ""
        return _filter


class Case(Field):
    def __init__(self, case):
        self._case = case

    def count(self):
        count_case_filter = pb.CountCaseFilter()
        if self._case.text.ByteSize():
            count_case_filter.text.CopyFrom(self._case.text)
        elif self._case.numeric.ByteSize():
            count_case_filter.numeric.CopyFrom(self._case.numeric)
        elif self._case.boolean.ByteSize():
            count_case_filter.boolean.CopyFrom(self._case.boolean)
        self._count_case = count_case_filter
        return self

    def _build(self, operator, value):
        _filter = pb.Filter()
        self._count_case.operator = operator
        self._count_case.value = value
        _filter.count_case.CopyFrom(self._count_case)
        return _filter

    def __gt__(self, value):
        return self._build("gt", value)

    def __lt__(self, value):
        return self._build("lt", value)

    def __ge__(self, value):
        return self._build("gte", value)

    def __le__(self, value):
        return self._build("lte", value)


class FilterBuilder(object):
    def __init__(self):
        self._filters = []

    def add(self, filter):
        self._filters.append(filter)
        return self

    def build(self):
        query = pb.Query()
        query.filters.extend(self._filters)
        return query


def convert_json_query_to_proto(json_query):
    query = pb.Query()
    return Parse(text=json_query, message=query)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ai/h2o/featurestore/api/v1/DashboardApi.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ai/h2o/featurestore/api/v1/DashboardApi.proto',
  package='ai.h2o.featurestore.api.v1',
  syntax='proto3',
  serialized_options=b'B\021DashboardApiProto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n-ai/h2o/featurestore/api/v1/DashboardApi.proto\x12\x1a\x61i.h2o.featurestore.api.v1\"a\n\x11\x44\x61shboardResponse\x12\x16\n\x0eproject_number\x18\x01 \x01(\x04\x12\x1b\n\x13\x66\x65\x61ture_sets_number\x18\x02 \x01(\x04\x12\x17\n\x0f\x66\x65\x61tures_number\x18\x03 \x01(\x04\x42\x13\x42\x11\x44\x61shboardApiProtob\x06proto3'
)




_DASHBOARDRESPONSE = _descriptor.Descriptor(
  name='DashboardResponse',
  full_name='ai.h2o.featurestore.api.v1.DashboardResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project_number', full_name='ai.h2o.featurestore.api.v1.DashboardResponse.project_number', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_sets_number', full_name='ai.h2o.featurestore.api.v1.DashboardResponse.feature_sets_number', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='features_number', full_name='ai.h2o.featurestore.api.v1.DashboardResponse.features_number', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=77,
  serialized_end=174,
)

DESCRIPTOR.message_types_by_name['DashboardResponse'] = _DASHBOARDRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DashboardResponse = _reflection.GeneratedProtocolMessageType('DashboardResponse', (_message.Message,), {
  'DESCRIPTOR' : _DASHBOARDRESPONSE,
  '__module__' : 'ai.h2o.featurestore.api.v1.DashboardApi_pb2'
  # @@protoc_insertion_point(class_scope:ai.h2o.featurestore.api.v1.DashboardResponse)
  })
_sym_db.RegisterMessage(DashboardResponse)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from typing import Iterator, List

from google.protobuf.empty_pb2 import Empty

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..collections.feature_sets import FeatureSets
from ..entities.feature_set import FeatureSet
from ..entities.project import Project


class Projects:
    def __init__(self, stub):
        self._stub = stub
        self._default_lock_value = None

    def list(self) -> Iterator[Project]:
        request = pb.ListProjectsPageRequest()
        while request:
            response = self._stub.ListProjectsPage(request)
            if response.next_page_token:
                request = pb.ListProjectsPageRequest()
                request.page_token = response.next_page_token
            else:
                request = None
            for project in response.projects:
                yield Project(self._stub, project)

    def list_feature_sets(self, project_names=[], filters=None) -> Iterator[FeatureSet]:
        request = pb.ListFeatureSetsPageRequest()
        request.project_names.extend(project_names)
        if filters:
            query = FeatureSets._build_feature_set_list_query(None, filters)
            request.query.CopyFrom(query)
        while request:
            response = self._stub.ListFeatureSetsPage(request)
            if response.next_page_token:
                request.page_token = response.next_page_token
            else:
                request = None
            for feature_set in response.feature_sets:
                yield FeatureSet(self._stub, feature_set)

    def create(
        self,
        project_name: str,
        description: str = "",
        secret: bool = False,
        locked: bool = None,
    ) -> Project:

        if locked is None and self._default_lock_value is None:
            self._default_lock_value = self._stub.GetProjectsDefault(Empty()).locked
        if locked is None:
            locked_value = self._default_lock_value
        else:
            locked_value = locked
        request = pb.CreateProjectRequest()
        request.secret = secret
        request.project_name = project_name
        request.description = description
        request.locked = locked_value
        response = self._stub.CreateProject(request)
        if response.already_exists:
            print("Project '" + project_name + "' already exists.")
        return Project(self._stub, response.project)

    def get(self, project_name: str) -> Project:
        request = pb.GetProjectRequest()
        request.project_name = project_name
        response = self._stub.GetProject(request)
        return Project(self._stub, response.project)

    def __repr__(self):
        return "This class wraps together methods working with projects"



from codecs import open
from os import path

from featurestore.client import Client
from featurestore.core.client_config import ClientConfig
from featurestore.core.collections.classifiers import (
    EmptyClassifier,
    RegexClassifier,
    SampleClassifier,
)
from featurestore.core.data_source_wrappers import (
    CSVFile,
    CSVFolder,
    DeltaTable,
    DeltaTableFilter,
    JdbcTable,
    JSONFile,
    JSONFolder,
    ParquetFile,
    ParquetFolder,
    PartitionOptions,
    Proxy,
    SnowflakeCursor,
    SnowflakeTable,
    SparkDataFrame,
)
from featurestore.core.schema import FeatureSchema, Schema
from featurestore.core.transformations import (
    DriverlessAIMOJO,
    JoinFeatureSets,
    SparkPipeline,
)
from featurestore.core.user_credentials import (
    AzureKeyCredentials,
    AzurePrincipalCredentials,
    AzureSasCredentials,
    PostgresCredentials,
    S3Credentials,
    SnowflakeCredentials,
    TeradataCredentials,
)


def __get_version():
    here = path.abspath(path.dirname(__file__))
    with open(path.join(here, "version.txt"), encoding="utf-8") as f:
        return f.read().strip()


__version__ = __get_version()
__all__ = [
    "Client",
    "FeatureSchema",
    "Schema",
    "CSVFile",
    "CSVFolder",
    "JSONFile",
    "JSONFolder",
    "ParquetFile",
    "ParquetFolder",
    "SnowflakeTable",
    "JdbcTable",
    "PartitionOptions",
    "SnowflakeCursor",
    "DeltaTable",
    "DeltaTableFilter",
    "Proxy",
    "JSONFolder",
    "Schema",
    "SparkDataFrame",
    "ClientConfig",
    "AzureKeyCredentials",
    "AzureSasCredentials",
    "AzurePrincipalCredentials",
    "S3Credentials",
    "TeradataCredentials",
    "SnowflakeCredentials",
    "PostgresCredentials",
    "SparkPipeline",
    "DriverlessAIMOJO",
    "JoinFeatureSets",
    "EmptyClassifier",
    "RegexClassifier",
    "SampleClassifier",
]



import abc
import collections
import sys
import time
import uuid
from random import randint
from typing import *

import grpc

from .auth import AuthException


class _ClientCallDetails(
    collections.namedtuple(
        "_ClientCallDetails", ("method", "timeout", "metadata", "credentials")
    ),
    grpc.ClientCallDetails,
):
    pass


class AuthClientInterceptor(
    grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor
):
    def __init__(self, client):
        self._client = client

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):
        if client_call_details.method in (
            "/ai.h2o.featurestore.api.v1.CoreService/Login",
            "/ai.h2o.featurestore.api.v1.CoreService/GetAccessToken",
            "/ai.h2o.featurestore.api.v1.CoreService/GetVersion",
        ):
            return continuation(client_call_details, request_or_iterator)
        else:
            metadata = []
            if client_call_details.metadata is not None:
                metadata = list(client_call_details.metadata)
            is_success, token_or_error = self._client.auth._obtain_token()
            if is_success:
                sys.tracebacklimit = None
                metadata.append(
                    (
                        "authorization",
                        token_or_error,
                    )
                )
            else:
                sys.tracebacklimit = 0
                raise AuthException(token_or_error)
            metadata.append(("request-id", str(uuid.uuid4())))
            new_client_call_details = _ClientCallDetails(
                client_call_details.method,
                self._client._client_config.timeout,
                metadata,
                client_call_details.credentials,
            )
            return continuation(new_client_call_details, request_or_iterator)

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(
        self, continuation, client_call_details, request_iterator
    ):
        return self._intercept_call(continuation, client_call_details, request_iterator)


class SleepingPolicy(abc.ABC):
    @abc.abstractmethod
    def sleep(self, try_i: int):
        """
        How long to sleep in milliseconds.
        :param try_i: the number of retry (starting from zero)
        """
        assert try_i >= 0


class ExponentialBackoff(SleepingPolicy):
    def __init__(self, *, init_backoff_ms: int, max_backoff_ms: int, multiplier: int):
        self.init_backoff = randint(0, init_backoff_ms)
        self.max_backoff = max_backoff_ms
        self.multiplier = multiplier

    def sleep(self, try_i: int):
        sleep_range = min(
            self.init_backoff * self.multiplier**try_i, self.max_backoff
        )
        sleep_ms = randint(0, sleep_range)
        time.sleep(sleep_ms / 1000)


class RetryOnRpcErrorClientInterceptor(
    grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor
):
    def __init__(
        self,
        *,
        max_attempts: int,
        sleeping_policy: SleepingPolicy,
        status_for_retry: Optional[Tuple[grpc.StatusCode]] = None,
    ):
        self.max_attempts = max_attempts
        self.sleeping_policy = sleeping_policy
        self.status_for_retry = status_for_retry

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):

        for try_i in range(self.max_attempts):
            response = continuation(client_call_details, request_or_iterator)

            if isinstance(response, grpc.RpcError):

                # Return if it was last attempt
                if try_i == (self.max_attempts - 1):
                    return response

                # If status code is not in retryable status codes
                if (
                    self.status_for_retry
                    and response.code() not in self.status_for_retry
                ):
                    return response

                self.sleeping_policy.sleep(try_i)
            else:
                return response

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(
        self, continuation, client_call_details, request_iterator
    ):
        return self._intercept_call(continuation, client_call_details, request_iterator)
